import asyncio
import copy
import logging
import re
import time
from dataclasses import dataclass
from enum import Enum

from .error import ParsingError


def crc(s: bytes) -> int:
    """Calculate a CRC as accepted by Simarine devices.

    Original source: https://github.com/htool/pico2signalk
    Copyright Erik Bosman / @brainsmoke
    """
    poly = 0x1189
    crc = 0
    for c in s:
        for i in range(8):
            c_msb = (c >> 7) & 1
            crc_msb = (crc >> 15) & 1

            c = (c << 1) & 0xFF
            crc = (crc << 1) & 0xFFFF

            if c_msb ^ crc_msb:
                crc ^= poly
    return crc


class MessageType(Enum):
    """The message type described by the header.
    The enum only represents the known subset.
    """

    DEVICE_COUNT = b"\x02"
    """Request the number of connected devices"""

    DEVICE_INFO = b"\x41"
    """Request information about a device"""

    SENSOR_STATE = b"\xb0"
    """Sensor update message (UDP)"""


@dataclass
class Header:
    """Describes the header of a message"""

    type: bytes
    length: int


HEADER_LENGTH = 14


@dataclass
class Message:
    """A message as received by a Simarine device"""

    type: MessageType
    data: bytes = b""


def parse_header(data: bytes) -> Header:
    """Parses the given bytes. Returns a Header on success.
    Raises ParsingError if the given bytes are not a valid header.
    Note: The header is not fully understood and might not work on
    all Simarine devices.
    """

    if match := re.match(
        b"\x00\x00\x00\x00\x00\xff(.)....(..)\xff", data[:HEADER_LENGTH]
    ):
        return Header(type=match[1], length=int.from_bytes(match[2]))
    else:
        raise ParsingError(f"Invalid header {data[:HEADER_LENGTH]!r}")


def parse_response(raw_response: bytes) -> Message:
    """Parse a response from a Simarine device. Returns a Message on success.
    Raises ParsingError if the given data is not a valid or known Simarine Message.
    """

    header = parse_header(raw_response[:HEADER_LENGTH])
    data_length = len(raw_response[HEADER_LENGTH - 1 :])

    if header.length != data_length:
        raise ParsingError(
            f"Incomplete response ({data_length} of {header.length} bytes received)",
        )

    calculated_crc = crc(raw_response[1:-3]).to_bytes(2)
    received_crc = raw_response[-2:]
    if calculated_crc != received_crc:
        raise ParsingError(f"Invalid crc ({received_crc!r} != {calculated_crc!r})")

    try:
        return Message(
            type=MessageType(header.type), data=raw_response[HEADER_LENGTH:-2]
        )
    except ValueError:
        raise ParsingError(f"Unknown message type {header.type.hex()}")


def make_request(message: Message) -> bytes:
    """Returns a valid Simarine request for the given Message."""

    length = 3 + len(message.data)
    header = b"\x00\x00\x00\x00\x00\xff%b\x04\x8c\x55\x4b%b\xff"
    data = header % (message.type.value, length.to_bytes(2)) + message.data
    return data + crc(data[1:-1]).to_bytes(2)


def make_device_request(device_id: int) -> Message:
    """Returns a device request for the given Device ID. The Simarine device
    will reply with a device response that contains the description of the
    device."""

    data = b"\x00\x01\x00\x00\x00%b\xff\x01\x03\x00\x00\x00\x00\xff\x00\x00\x00\x00\xff"
    return Message(
        type=MessageType.DEVICE_INFO,
        data=data % device_id.to_bytes(1),
    )


# UDP port used by Simarine devices to announce sensor changes
DEFAULT_UDP_PORT = 43210

# TCP port used by Simarine devices to exchange request/response messages
DEFAULT_TCP_PORT = 5001


Address = tuple[str, int]


class _UdpMessage:
    """Represents a UDP message. Used to share state between objects."""

    def __init__(self, data: bytes, addr: Address) -> None:
        self.data = data
        self.addr = addr


class _UDPServerProtocol(asyncio.DatagramProtocol):
    """Implements a UDP protocol that notifies the given event when a new UDP message
    is received and writes it to the given message."""

    def __init__(self, event: asyncio.Event, message: _UdpMessage) -> None:
        self.event = event
        self.message = message

    def connection_made(self, transport: asyncio.transports.BaseTransport) -> None:
        self.transport = transport

    def datagram_received(self, data: bytes, addr: Address) -> None:
        self.message.data = data
        self.message.addr = addr
        logging.debug("UDP broadcast message received")
        self.event.set()


class Communication:
    """Communicate with Simarine devices by making requests and listening
    for broadcast messages.

    Simarine devices communicate via TCP and UDP broadcast messages.
    UDP broadcast messages report new sensor values in short intervals and
    can be used to discover the device. TCP is used to request specific
    information and probably make changes to the device."""

    def __init__(self, udp_port=DEFAULT_UDP_PORT, tcp_port=DEFAULT_TCP_PORT) -> None:
        self.udp_port = udp_port
        self.tcp_port = tcp_port
        self.ip_address: str | None = None

        self._udp_event = asyncio.Event()
        self._latest_udp_message = _UdpMessage(b"", ("", 0))
        self._udp_transport: asyncio.transports.DatagramTransport | None = None
        self._last_request_time: float | None = None

        self._tcp_reader: asyncio.StreamReader | None = None
        self._tcp_writer: asyncio.StreamWriter | None = None

    async def create_udp_server(self):
        """Create an UDP broadcast server that listens to incoming messages"""

        if self._udp_transport is None:
            logging.debug("Start UDP broadcast server on port %s", self.udp_port)
            loop = asyncio.get_event_loop()
            self._udp_transport, _ = await loop.create_datagram_endpoint(
                lambda: _UDPServerProtocol(self._udp_event, self._latest_udp_message),
                local_addr=("0.0.0.0", self.udp_port),
                reuse_port=True,
            )

    async def discover_ip(self):
        """Find the Simarine device by waiting for a UDP broadcast message"""

        if self.ip_address is None:
            logging.info("Discovering IP of Simarine device...")

            udp_message = await self._receive_udp()
            self.ip_address, _ = udp_message.addr

            logging.info("Found device at %s:%s", self.ip_address, self.tcp_port)

    async def close(self) -> None:
        """Closes an open connection"""

        if self._tcp_writer is not None:
            logging.info("Disconnecting from Simarine device...")
            self._tcp_writer.close()
            await self._tcp_writer.wait_closed()

        if self._udp_transport is not None:
            self._udp_transport.close()
            self._udp_transport = None
            self.ip_address = None

    async def connect(self) -> None:
        logging.debug("Connecting to %s:%i...", self.ip_address, self.tcp_port)
        self._tcp_reader, self._tcp_writer = await asyncio.open_connection(
            self.ip_address, self.tcp_port
        )

    async def __aenter__(self):
        await self.create_udp_server()  # pragma: no cover

    async def __aexit__(self, exc_t, exc_v, exc_tb):
        await self.close()  # pragma: no cover

    async def request(self, message: Message) -> Message:
        """Send a request to a Simarine device over TCP and returns a response
        Raises ParsingError in case no valid response has been returned."""

        assert self.ip_address is not None
        assert self._tcp_reader is not None
        assert self._tcp_writer is not None

        try:
            logging.debug("Requesting %r", message.type.name)
            self._tcp_writer.write(make_request(message))
            await self._tcp_writer.drain()

            data = await self._tcp_reader.read(2048)
            response = parse_response(data)

        finally:
            self._last_request_time = time.time()

        return response

    async def receive_broadcast(self) -> Message:
        """Receive the next incoming UDP message
        Raises ParsingError in case the next message is invalid."""

        assert self._udp_transport is not None
        udp_message = await self._receive_udp()
        response = parse_response(udp_message.data)
        logging.debug("Received broadcast message %r", response.type.name)
        return response

    async def _receive_udp(self) -> _UdpMessage:
        """Waits until the next UDP broadcast message is received and returns it"""

        logging.debug("Waiting for UDP broadcast message")
        await self._udp_event.wait()
        message = copy.copy(self._latest_udp_message)
        self._udp_event.clear()
        return message



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, ClassVar, Type

from .property_dict import PropertyDict, PropertyValue


@dataclass
class Sensor(ABC):
    """Represents a sensor of a device"""

    UNIT: ClassVar[str] = ""
    """A string describing the unit of the value property for this sensor"""

    sensor_id: int | None = field(default=None, kw_only=True)
    """A virtual unique identifier of the sensor"""

    value: float = 0.0
    """The current value of the sensor"""

    state_index: int | None = field(
        default=None, repr=False, compare=False, kw_only=True
    )
    """The absolute index in a state PropertyDict"""

    state_offset: int = field(default=0, repr=False, compare=False, kw_only=True)
    """The offset of this sensor's state in a PropertyDict relative to the first
    sensor in a device. Used to define the sensor inside a device."""

    def update(self, property_dict: PropertyDict):
        """Update the sensor value based on a PropertyDict from a state update"""

        assert self.state_index is not None
        p = property_dict.values[self.state_index]
        self.value = self.convert(p)

    @abstractmethod
    def convert(self, p: PropertyValue) -> float:
        """Converts the given PropertyValue to the sensor value in the defined unit"""
        pass

    def asdict(self) -> dict[str, Any]:
        """Convert the sensor to a dict leaving out internal properties. Useful for
        converting a sensor to JSON."""

        assert self.sensor_id is not None
        return {
            "id": self.sensor_id,
            "value": self.value,
            "unit": self.UNIT,
        }


@dataclass
class BarometerSensor(Sensor):
    UNIT: ClassVar[str] = "mbar"

    def convert(self, p: PropertyValue) -> float:
        return p.number


@dataclass
class TemperatureSensor(Sensor):
    UNIT: ClassVar[str] = "celsius"

    def convert(self, p: PropertyValue) -> float:
        return p.second / 10.0


@dataclass
class CurrentSensor(Sensor):
    UNIT: ClassVar[str] = "ampere"

    def convert(self, p: PropertyValue) -> float:
        return p.second / 100.0


@dataclass
class ResistiveSensor(Sensor):
    UNIT: ClassVar[str] = "ohm"

    def convert(self, p: PropertyValue) -> float:
        return p.second


@dataclass
class VoltageSensor(Sensor):
    UNIT: ClassVar[str] = "volt"

    def convert(self, p: PropertyValue) -> float:
        return p.second / 1000.0


@dataclass
class TankVolumeSensor(Sensor):
    UNIT: ClassVar[str] = "liter"

    def convert(self, p: PropertyValue) -> float:
        return p.second / 10.0


@dataclass
class TankLevelSensor(Sensor):
    UNIT: ClassVar[str] = "percentage"

    def convert(self, p: PropertyValue) -> float:
        return p.first / 1000.0


@dataclass
class ChargeSensor(Sensor):
    UNIT: ClassVar[str] = "percentage"

    def convert(self, p: PropertyValue) -> float:
        return p.first / 16000.0


@dataclass
class CapacitySensor(Sensor):
    UNIT: ClassVar[str] = "ampere_hour"

    def convert(self, p: PropertyValue) -> float:
        return p.second / 100.0


def sensor_field(sensory_cls: Type[Sensor], state_offset=0):
    return field(default_factory=lambda: sensory_cls(state_offset=state_offset))



import logging
from typing import AsyncIterator

from .communication import (
    Communication,
    Message,
    MessageType,
    make_device_request,
)
from .device import Device, NullDevice, UnknownDevice, device_from_property_dict
from .error import ParsingError
from .property_dict import PropertyDict, parse_property_dict
from .sensor import Sensor


class DeviceReader:
    """
    Read the current state for devices in a Simarine setup.
    It has only been tested with the Simarine Pico rev2.

    Note that only one device can be connected at the same time including the Simarine app.

    open() will connect and read all device information. Afterwards devices are accessible
    via the devices property.

    read_sensors() will wait for a state update message and update the sensor values of all
    devices.
    """

    def __init__(self, communication: Communication | None = None) -> None:
        """Creates a new device reader"""

        self.communication = Communication() if communication is None else communication
        self.devices: list[Device] = []
        self.sensors: list[Sensor] = []

    async def open(self) -> None:
        """Opens the communication with a Simarine device and requests all device
        information or loads them from a file.
        """

        await self.communication.create_udp_server()
        await self.communication.discover_ip()
        await self.communication.connect()
        self._set_devices(list([device async for device in self._request_devices()]))

    def close(self):
        """Closes an open communication"""

        self.communication.close()

    async def read_sensors(self) -> None:
        """Waits for a broadcast message and updates all sensor values"""

        try:
            response = await self.communication.receive_broadcast()
            while response.type != MessageType.SENSOR_STATE:
                response = await self.communication.receive_broadcast()
            property_dict = parse_property_dict(response.data)

            for device in self.devices:
                device.update_sensors(property_dict)
        except ParsingError as e:
            logging.debug(e)

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_t, exc_v, exc_tb):
        self.close()

    async def _request_device_count(self) -> int:
        return _parse_device_count_response(
            await self.communication.request(Message(type=MessageType.DEVICE_COUNT))
        )

    async def _request_devices(self) -> AsyncIterator[Device]:
        logging.info("Fetching device infos")
        sensor_start_index = 0
        for device_id in range(await self._request_device_count()):
            device_response = await self.communication.request(
                make_device_request(device_id)
            )
            property_dict = _parse_device_response(device_response)
            device = device_from_property_dict(device_id, property_dict)
            device.initialize_sensor_indices(sensor_start_index)
            if not isinstance(device, (NullDevice, UnknownDevice)):
                yield device
            sensor_start_index += device.SENSOR_INDEX_OFFSET

    def _set_devices(self, devices):
        self.devices = devices
        self.sensors = [
            sensor
            for device in self.devices
            for _, sensor in device.sensors
            if sensor.sensor_id is not None
        ]


def _parse_device_count_response(response: Message) -> int:
    if response.type == MessageType.DEVICE_COUNT and len(response.data) >= 6:
        return response.data[5] + 1
    else:
        raise ParsingError(f"Couldn't get device count. Unexpected response {response}")


def _parse_device_response(response: Message) -> PropertyDict:
    if response.type == MessageType.DEVICE_INFO:
        return parse_property_dict(response.data)
    else:
        raise ParsingError(f"Couldn't get device info. Unexpected response {response}")



from dataclasses import dataclass, fields
from enum import Enum
from typing import Any, ClassVar, Iterable

from .property_dict import PropertyDict
from .sensor import (
    BarometerSensor,
    CapacitySensor,
    ChargeSensor,
    CurrentSensor,
    ResistiveSensor,
    Sensor,
    TankLevelSensor,
    TankVolumeSensor,
    TemperatureSensor,
    VoltageSensor,
    sensor_field,
)


class FluidType(Enum):
    FRESH_WATER = 1
    FUEL = 2
    WASTE_WATER = 3


class BatteryType(Enum):
    WET_LOW_MAINTENANCE = 1
    WET_MAINTENANCE_FREE = 2
    AGM = 3
    DEEP_CYCLE = 4
    GEL = 5
    LIFEPO4 = 6


@dataclass
class Device:
    """Represents a device in a Simarine setup.

    A device has a name, static properties describing the device and one or more sensors
    that are regularly updated."""

    TYPE: ClassVar[str] = ""
    """String describing the type of device"""

    SENSOR_INDEX_OFFSET: ClassVar[int] = 0
    """The size of the device inside a state update. Defines how many
    values have to be skipped to get to the next device."""

    device_id: int
    """A unique identifier of the device as provided by the Simarine device"""

    name: str
    """The name of the device as configured by the user"""

    @property
    def sensors(self) -> Iterable[tuple[str, Sensor]]:
        """Iterator for all sensors and their property name of this device"""

        for f in fields(self):
            attr = getattr(self, f.name)
            if isinstance(attr, Sensor):
                yield f.name, attr

    def update_sensors(self, property_dict: PropertyDict) -> None:
        """Update all sensors for this device. Should be called when a state update
        has been received."""

        for _, sensor in self.sensors:
            sensor.update(property_dict)

    def asdict(self) -> dict[str, Any]:
        """Convert the device to a dict leaving out internal properties. Useful for
        converting a device to JSON."""

        def convert_value(obj):
            if isinstance(obj, Enum):
                return obj.name.lower()
            if isinstance(obj, Sensor):
                return obj.asdict()
            return obj

        d: dict = {f.name: convert_value(getattr(self, f.name)) for f in fields(self)}

        d["type"] = self.TYPE

        return d

    def initialize_sensor_indices(self, state_start_index: int) -> None:
        """Needs to be called after a device has been constructed to initialize
        all its sensors. This is cecessary for processing state updates."""

        for index, (_, sensor) in enumerate(self.sensors):
            sensor.state_index = state_start_index + sensor.state_offset
            # We have "virtual" sensors that might have the same state_index so let's
            # assign a unique index to each sensor. To avoid ids overlapping with the next
            # device we scale the available range
            sensor.sensor_id = state_start_index * 100 + index


@dataclass
class PicoInternal(Device):
    TYPE: ClassVar[str] = "pico_internal"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 6

    voltage: VoltageSensor = sensor_field(VoltageSensor)


@dataclass
class Battery(Device):
    TYPE: ClassVar[str] = "battery"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 5

    capacity: float
    battery_type: BatteryType

    charge: ChargeSensor = sensor_field(ChargeSensor, 0)
    remaining_capacity: CapacitySensor = sensor_field(CapacitySensor, 0)
    current: CurrentSensor = sensor_field(CurrentSensor, 1)
    voltage: VoltageSensor = sensor_field(VoltageSensor, 2)


@dataclass
class Tank(Device):
    TYPE: ClassVar[str] = "tank"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 1

    capacity: float
    fluid_type: FluidType

    volume: TankVolumeSensor = sensor_field(TankVolumeSensor, 0)
    level: TankLevelSensor = sensor_field(TankLevelSensor, 0)


@dataclass
class TemperatureDevice(Device):
    TYPE: ClassVar[str] = "temperature"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 1

    temperature: TemperatureSensor = sensor_field(TemperatureSensor)


@dataclass
class Barometer(Device):
    TYPE: ClassVar[str] = "barometer"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 2

    pressure: Sensor = sensor_field(BarometerSensor)


@dataclass
class VoltageDevice(Device):
    TYPE: ClassVar[str] = "voltage"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 1

    voltage: VoltageSensor = sensor_field(VoltageSensor)


@dataclass
class CurrentDevice(Device):
    TYPE: ClassVar[str] = "current"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 2

    current: CurrentSensor = sensor_field(CurrentSensor)


@dataclass
class ResistiveDevice(Device):
    TYPE: ClassVar[str] = "resistive"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 1

    resistance: Sensor = sensor_field(ResistiveSensor)


@dataclass
class NullDevice(Device):
    TYPE: ClassVar[str] = "null"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 0


@dataclass
class UnknownDevice(Device):
    TYPE: ClassVar[str] = "unknown"
    SENSOR_INDEX_OFFSET: ClassVar[int] = 1


def device_from_property_dict(device_id: int, property_dict: PropertyDict) -> Device:
    """Converts a PropertyDict describing a device into the corresponding Device object"""

    device_type = property_dict.values[1][1]

    if device_type == 0:
        return NullDevice(device_id=device_id, name="<null>")
    elif device_type == 1:
        if property_dict.strings[3] == "PICO INTERNAL":
            return PicoInternal(device_id=device_id, name=property_dict.strings[3])
        else:
            return VoltageDevice(device_id=device_id, name=property_dict.strings[3])
    elif device_type == 2:
        return CurrentDevice(device_id=device_id, name=property_dict.strings[3])
    elif device_type == 3:
        return TemperatureDevice(device_id=device_id, name=property_dict.strings[3])
    elif device_type == 5:
        return Barometer(device_id=device_id, name=property_dict.strings[3])
    elif device_type == 6:
        return ResistiveDevice(device_id=device_id, name=property_dict.strings[3])
    elif device_type == 8:
        return Tank(
            device_id=device_id,
            name=property_dict.strings[3],
            capacity=property_dict.values[7][1] / 10.0,
            fluid_type=FluidType(property_dict.values[6][1]),
        )
    elif device_type == 9:
        return Battery(
            device_id=device_id,
            name=property_dict.strings[3],
            capacity=property_dict.values[5][1] / 100.0,
            battery_type=BatteryType(property_dict.values[8][1]),
        )

    return UnknownDevice(device_id=device_id, name="<unknown>")



from .communication import Communication
from .device import (
    Barometer,
    Battery,
    BatteryType,
    CurrentDevice,
    Device,
    FluidType,
    PicoInternal,
    ResistiveDevice,
    Tank,
    TemperatureDevice,
    VoltageDevice,
)
from .device_reader import DeviceReader
from .error import ParsingError
from .sensor import (
    BarometerSensor,
    CapacitySensor,
    ChargeSensor,
    CurrentSensor,
    ResistiveSensor,
    Sensor,
    TankLevelSensor,
    TankVolumeSensor,
    TemperatureSensor,
    VoltageSensor,
    sensor_field,
)

__all__: tuple[str, ...] = (
    "Barometer",
    "BarometerSensor",
    "Battery",
    "BatteryType",
    "CapacitySensor",
    "ChargeSensor",
    "Communication",
    "CurrentDevice",
    "CurrentSensor",
    "Device",
    "DeviceReader",
    "FluidType",
    "ParsingError",
    "PicoInternal",
    "ResistiveDevice",
    "ResistiveSensor",
    "sensor_field",
    "Sensor",
    "Tank",
    "TankLevelSensor",
    "TankVolumeSensor",
    "TemperatureDevice",
    "TemperatureSensor",
    "VoltageDevice",
    "VoltageSensor",
)
